"""
Broadworks OCI-P Interface Base Classes

Base classes used by the types, requests and responses as well as
other components like ElementInfo that are used by those.
"""
import logging
import re
from collections import namedtuple
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Tuple

import attr
from lxml import etree
from null_object import Null

from broadworks_ocip.exceptions import OCIErrorAPISetup
from broadworks_ocip.exceptions import OCIErrorAttributeMissing
from broadworks_ocip.exceptions import OCIErrorResponse
from broadworks_ocip.exceptions import OCIErrorUnexpectedAttribute


logger = logging.getLogger(__name__)

# constants
XML_NIL = "{http://www.w3.org/2001/XMLSchema-instance}nil"
XSD_INST = "http://www.w3.org/2001/XMLSchema-instance"
XSD_TYPE = "{http://www.w3.org/2001/XMLSchema-instance}type"


@attr.s(slots=True, frozen=True)
class ElementInfo:
    """
    ElementInfo - information on each element of a Broadsoft OCIType

    Used to describe the element when serialising to/from XML

    Attributes:
        name: name of this element (in python snake case)
        xmlname: name of this element (in Java like camel case)
        type: the type of the resulting element
        is_complex: is this a complex element - containing another OCIType derived class
        is_required: Is this required (True) or Optional (False)
        is_array: Is this an array/list of element values
        is_table: Is this a Broadworks table type - only seen in Responses
    """

    name: str = attr.ib()
    xmlname: str = attr.ib()
    type = attr.ib()
    is_complex: bool = attr.ib(default=False)
    is_required: bool = attr.ib(default=False)
    is_array: bool = attr.ib(default=False)
    is_table: bool = attr.ib(default=False)
    is_abstract: bool = attr.ib(default=False)
    is_container: bool = attr.ib(default=False)


class OCIType:
    """
    OCIType - Base type for all the OCI-P component classes

    There are no attributes of this base class (the `_frozen` attribute is
    used as a flag to lock the instance).  The attributes are added in the
    various subclasses.
    """

    __slots__: List[str] = ["_frozen"]

    def __init__(self, **kwargs):
        cname = type(self).__name__  # needed for error messages
        for elem in self._elements():
            if elem.name in kwargs:
                value = kwargs[elem.name]
                if elem.is_required and value is None:
                    raise OCIErrorAttributeMissing(
                        message=f"{cname}: Required attribute {elem.name} is missing",
                    )
                if value is Null:
                    # always allowed
                    pass
                elif elem.is_array:
                    if not isinstance(value, list):
                        raise TypeError(
                            f"{cname}: Expected {elem.name} to be a list but it is {type(value)}",
                        )
                    if len(value) > 0 and not isinstance(value[0], elem.type):
                        raise TypeError(
                            f"{cname}: Expected first element of {elem.name} to be type {elem.type} but it is {type(value[0])}",
                        )
                elif elem.is_table:
                    if not isinstance(value, list):
                        raise TypeError(
                            f"{cname}: Expected {elem.name} to be a table/list but it is {type(value)}",
                        )
                elif elem.is_container:
                    if value is not None and not isinstance(value, dict):
                        raise TypeError(
                            f"{cname}: Expected {elem.name} to be a dict of elements but it is {type(value)}",
                        )
                elif not isinstance(value, elem.type):
                    raise TypeError(
                        f"{cname}: Expected {elem.name} to be type {elem.type} but it is {type(value)}",
                    )
                setattr(self, elem.name, value)
                del kwargs[elem.name]
            elif elem.is_required:
                raise OCIErrorAttributeMissing(
                    message=f"{cname}: Required attribute {elem.name} is missing",
                )
            else:
                setattr(self, elem.name, None)
        if kwargs:
            raise OCIErrorUnexpectedAttribute(
                message=f"{cname}: Unexpected attribute(s) {kwargs.keys()}",
            )
        self._frozen = True

    def __delattr__(self, *args, **kwargs):
        if hasattr(self, "_frozen"):
            raise AttributeError("This object is frozen!")
        object.__delattr__(self, *args, **kwargs)

    def __setattr__(self, *args, **kwargs):
        if hasattr(self, "_frozen"):
            raise AttributeError("This object is frozen!")
        object.__setattr__(self, *args, **kwargs)

    # Namespace maps used for various XML build tasks
    @classmethod
    def _default_nsmap(cls):
        return {None: "", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}

    @classmethod
    def _document_nsmap(cls):
        return {None: "C", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}

    @classmethod
    def _error_nsmap(cls):
        return {
            "c": "C",
            None: "",
            "xsi": "http://www.w3.org/2001/XMLSchema-instance",
        }

    @classmethod
    def _elements(cls) -> Tuple[ElementInfo, ...]:
        raise OCIErrorAPISetup(message="_elements should be defined in the subclass.")

    @property
    def type_(self):
        """Return the typename of the class"""
        return self.__class__.__name__

    @classmethod
    def class_to_property_(self, name):
        """Map a XML class name to the associated property"""
        return name[0].lower() + name[1:]

    def post_xml_decode_(self):
        """
        Carry out any operations after the XML decode

        Intended for use by subclasses where they need to take actions immediately
        after they are created from an incoming XML document.
        """
        pass

    def etree_components_(self, name=None):
        """
        Build XML etree element tree for this OCIType

        Arguments:
            name: The name or tag of the element - defaults to the `type_`

        Returns:
            etree: etree.Element() for this class
        """
        if name is None:
            name = self.type_
        element = etree.Element(name, nsmap=self._default_nsmap())
        return self.etree_sub_components_(element)

    def etree_sub_components_(self, element: etree._Element) -> etree._Element:
        """
        Build XML etree subelements for the components within this OCIType

        Arguments:
            element: The parent element that the components are to be attached to

        Returns:
            etree: etree.Element() for this class
        """
        for sub_element in self._elements():
            value = getattr(self, sub_element.name)
            if sub_element.is_array:
                if value is not None:
                    if len(value) == 0:
                        element.attrib[XML_NIL] = "true"
                    else:
                        for subvalue in value:
                            self.etree_sub_element_(element, sub_element, subvalue)
            else:
                self.etree_sub_element_(element, sub_element, value)
        return element

    def etree_sub_element_(
        self,
        element: etree._Element,
        sub_element: ElementInfo,
        value,
    ) -> etree._Element:
        """
        Build XML etree subelement for one element within this OCIType

        Arguments:
            element: The parent element that the components are to be attached to
            sub_element: The definition of the sub element to be attached
            value: Value of the sub element - quite possibly None

        Returns:
            etree: etree.Element() for this class
        """
        # if value is Null or (value is None and sub_element.is_required):
        if value is Null or (value is None and sub_element.is_required):
            etree.SubElement(
                element,
                sub_element.xmlname,
                {XML_NIL: "true"},
                nsmap=self._default_nsmap(),
            )
        elif value is None:
            pass
        elif sub_element.is_table:
            # any table should be a list of namedtuple elements
            if type(value) is list and len(value) > 0:
                elem = etree.SubElement(
                    element,
                    sub_element.xmlname,
                    nsmap=self._default_nsmap(),
                )
                first = value[0]
                for col in first._fields:
                    col_heading = etree.SubElement(elem, "colHeading")
                    col_heading.text = self.snake_case_to_column_header(col)
                for row in value:
                    row_item = etree.SubElement(elem, "row")
                    for col in row:
                        col_item = etree.SubElement(row_item, "col")
                        col_item.text = col
        elif sub_element.is_container:
            elem = etree.SubElement(
                element,
                sub_element.xmlname,
                nsmap=self._default_nsmap(),
            )
            for item in sub_element.type:
                if item.name in value:
                    self.etree_sub_element_(elem, item, value[item.name])
        elif sub_element.is_complex:
            if sub_element.is_abstract:
                elem = etree.SubElement(
                    element,
                    sub_element.xmlname,
                    {XSD_TYPE: value.type_},
                    nsmap=self._default_nsmap(),
                )
            else:
                elem = etree.SubElement(
                    element,
                    sub_element.xmlname,
                    nsmap=self._default_nsmap(),
                )
            value.etree_sub_components_(elem)
        else:
            elem = etree.SubElement(
                element,
                sub_element.xmlname,
                nsmap=self._default_nsmap(),
            )
            if sub_element.type == bool:
                elem.text = "true" if value else "false"
            elif sub_element.type == int:
                elem.text = str(value)
            else:
                elem.text = value
        return element

    @classmethod
    def column_header_snake_case_(cls, header: str) -> str:
        """
        Converts an XML name into a pythonic snake case name

        Arguments:
            header: The header name in space separated words

        Returns:
            snake: lower cased and underscore separated result name
        """
        return re.sub("[^A-Za-z0-9]+", r"_", header).lower()

    def snake_case_to_column_header(self, snake_str: str) -> str:
        """
        Converts a pythonic snake case name into a column header name

        Arguments:
            snake_str: The header name in snake lower case

        Returns:
            column_header: initial capital and space separated result name
        """
        components = snake_str.split("_")
        # We capitalize the first letter of each component except the first one
        # with the 'title' method and join them together.
        return " ".join(x.title() for x in components)

    @classmethod
    def decode_table_(cls, element: etree._Element) -> List[NamedTuple]:
        """
        Decode a table (used in a OCIResponse) into a list of named tuples

        Arguments:
            element: The OCITable XML element

        Returns:
            results: List of namedtuple elements, one for each table row
        """
        typename: str = element.tag
        results: List[NamedTuple] = []
        columns = [
            cls.column_header_snake_case_(b.text)
            for b in element.iterfind("colHeading")
        ]
        type: NamedTuple = namedtuple(typename, columns)  # type: ignore
        for row in element.iterfind("row"):
            rowdata = [b.text for b in row.iterfind("col")]
            rowobj = type(*rowdata)
            results.append(rowobj)
        return results

    @classmethod
    def build_from_etree_non_parameters_(
        cls,
        element: etree._Element,
        initialiser: dict,
    ) -> None:
        """
        Handle any items outside the parameter set

        Intended for use by subclasses where they need to take actions immediately
        after they are created from an incoming XML document.
        """
        pass

    @classmethod
    def build_from_node_(
        cls,
        api: "BroadworksAPI",  # type: ignore # noqa
        elem: ElementInfo,
        node: etree._Element,
    ) -> Any:
        """
        Creates an OCI subelement from a single  XML etree node

        Arguments:
            elem: The subelement descriptor
            node: The OCITable XML element node

        Returns:
            results: Object instance for this class
        """
        if node is not None:
            if XML_NIL in node.attrib and node.attrib[XML_NIL] == "true":
                return Null
            elif elem.is_table:
                return cls.decode_table_(node)
            elif elem.is_complex:
                if elem.is_abstract:
                    if XSD_TYPE in node.attrib:
                        thisclass = api.get_type_class(
                            node.attrib[XSD_TYPE],
                        )
                        logger.error(f"thisclass={thisclass}")
                        return thisclass.build_from_etree_(api=api, element=node)
                    else:
                        return ValueError("Cannot decode abstract object")
                elif elem.is_container:
                    return cls.build_initialiser_from_etree_(
                        api=api,
                        element=node,
                        subelement_set=elem.type,
                    )
                else:
                    return elem.type.build_from_etree_(api=api, element=node)
            elif elem.type == bool:
                return elem.type(
                    True if node.text == "true" else False,
                )
            else:
                return elem.type(node.text)
        else:
            return None

    @classmethod
    def build_initialiser_from_etree_(
        cls,
        api: "BroadworksAPI",  # type: ignore # noqa
        element: etree._Element,
        subelement_set: List[ElementInfo],
        extras: Dict[str, Any] = {},
    ):
        initialiser = extras.copy()
        for elem in subelement_set:
            if elem.is_array:
                result = []
                nodes = element.findall(elem.xmlname)
                for node in nodes:
                    result.append(cls.build_from_node_(api=api, elem=elem, node=node))
                initialiser[elem.name] = result
            else:
                node = element.find(elem.xmlname)
                if node is not None:
                    initialiser[elem.name] = cls.build_from_node_(
                        api=api,
                        elem=elem,
                        node=node,
                    )
                # else:
                # I am inclined to thow an error here - at least after checking if
                # the thing is require, but the class builder should do that so lets
                # let it do its thing
        return initialiser

    @classmethod
    def build_from_etree_(
        cls,
        api: "BroadworksAPI",  # type: ignore # noqa
        element: etree._Element,
        extras: Dict[str, Any] = {},
    ):
        """
        Create an OciType based instance from an XML etree element

        Arguments:
            element: The OCIType XML element

        Returns:
            results: Object instance for this class
        """
        initialiser = cls.build_initialiser_from_etree_(
            api=api,
            element=element,
            subelement_set=list(cls._elements()),
            extras=extras,
        )
        # now have a dict with all the bits in it.
        # use that to build a new object
        return cls(**initialiser)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert object to dict representation of itself

        This was provided as part of the Classforge system, which we have moved away
        from, so this is a local re-implementation.  This is only used within the test
        suite at present.
        """
        elements = {}
        for elem in self._elements():
            value = getattr(self, elem.name)
            if elem.is_table:
                pass
            elif elem.is_complex:
                if elem.is_array:
                    value = [x.to_dict() for x in value]
                elif value is not None:
                    value = value.to_dict()
            elif elem.is_array:
                value = [x for x in value]
            elements[elem.name] = value
        return elements

    def __repr__(self) -> str:
        """Convert object to string representation of itself."""
        dict_form = self.to_dict()
        bits = [f"{key}={repr(dict_form[key])}" for key in dict_form.keys()]
        return f"{self.__class__.__name__}({', '.join(bits)})"

    def __str__(self) -> str:
        """Convert object to string representation of itself."""
        return repr(self)


class OCICommand(OCIType):
    """
    OCICommand - base class for all OCI Command (Request/Response) types

    Attributes:
        session_id (str): The session ID used for the command exchange.
            We use UUIDs by default, although this is not required.  The
            default is fixed on here (but normally passed in from the
            containing API object) - do not use the default in production -
            its simply there to give a known value for testing.
    """

    __slots__: List[str] = ["session_id"]

    def __init__(
        self,
        session_id: str = "00000000-1111-2222-3333-444444444444",
        **kwargs,
    ):
        self.session_id = session_id
        super().__init__(**kwargs)

    def build_xml_(self) -> bytes:
        """
        Build an XML document of the current Command (Request/Response)

        Returns:
            xml: byte string containing XML document
        """
        # document root element
        root = etree.Element(
            "{C}BroadsoftDocument",
            {"protocol": "OCI"},
            nsmap=self._document_nsmap(),
        )
        #
        # add the session
        session = etree.SubElement(root, "sessionId", nsmap=self._default_nsmap())
        session.text = self.session_id
        #
        # and the command
        element = self.build_xml_command_element_(root)
        self.etree_sub_components_(element)  # attach parameters etc
        #
        # wrap a tree around it
        tree = etree.ElementTree(root)
        return etree.tostring(
            tree,
            xml_declaration=True,
            encoding="ISO-8859-1",
            # standalone=False,
            # pretty_print=True,
        )

    def build_xml_command_element_(self, root: etree._Element) -> etree._Element:
        """
        Build the XML etree of the main command element of the current Command
        Intended to be overridden in a subclass for the few elements that do things
        a little differently (for example errors).

        Returns:
            wrapped_element: The wrapped commane element tree
        """
        return etree.SubElement(
            root,
            "command",
            {XSD_TYPE: self.type_},
            nsmap=self._default_nsmap(),
        )

    @classmethod
    def build_from_etree_non_parameters_(
        cls,
        element: etree._Element,
        initialiser: dict,
    ):
        """
        Pick up the session id from the command set

        Overrides the class method defined in OCIType.
        """
        node = element.find("sessionId")
        if node is not None:
            initialiser["session_id"] = node.text

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert object to dict representation of itself

        This was provided as part of the Classforge system, which we have moved away
        from, so this is a local re-implementation.  This is only used within the test
        suite at present.
        """
        elements = super().to_dict()  # pick up the base object data
        return {"session_id": self.session_id, **elements}


class OCIRequest(OCICommand):
    """
    OCIRequest - base class for all OCI Command Request types
    """

    pass


class OCIResponse(OCICommand):
    """
    OCIResponse - base class for all OCI Command Response types
    """

    def build_xml_command_element_(self, root: etree._Element) -> etree._Element:
        """
        Build the XML etree of the main command element of the current Command
        Responses have an echo attribute in the element.

        Returns:
            etree: The XML etree of the main command element
        """
        return etree.SubElement(
            root,
            "command",
            {"echo": "", XSD_TYPE: self.type_},
            nsmap=self._default_nsmap(),
        )


class SuccessResponse(OCIResponse):
    """
    The SuccessResponse is concrete response sent whenever a transaction is successful
    and does not return any data.
    """

    __slots__: List[str] = []

    @classmethod
    def _elements(cls) -> Tuple[ElementInfo, ...]:
        return ()


class ErrorResponse(OCIResponse):
    """
    The ErrorResponse is concrete response sent whenever a transaction fails
    and does not return any data.

    As this an error, when it is created from an incoming command response, a
    `OCIErrorResponse` exception is raised in `post_xml_decode_`.

    Attributes:
        error_code (int): The error code
        summary (str): Summary of the error
        summary_english (str): Summary of the error in english!
        detail (str): Detail of the error
        type (str): Type of the error

    """

    __slots__: List[str] = [
        "error_code",
        "summary",
        "summary_english",
        "detail",
        "type",
    ]

    @classmethod
    def _elements(cls) -> Tuple[ElementInfo, ...]:
        return (
            ElementInfo("error_code", "errorCode", int),
            ElementInfo("summary", "summary", str, is_required=True),
            ElementInfo("summary_english", "summaryEnglish", str, is_required=True),
            ElementInfo("detail", "detail", str),
            ElementInfo("type", "type", str),
        )

    def post_xml_decode_(self):
        """Raise an exception as this is an error"""
        raise OCIErrorResponse(
            object=self,
            message=f"{self.error_code}: {self.summary} - {self.detail}",
        )

    def build_xml_command_element_(self, root):
        return etree.SubElement(
            root,
            "command",
            {
                "type": "Error",
                "echo": "",
                XSD_TYPE: "c:" + self.type_,
            },
            nsmap=self._error_nsmap(),
        )


# end



#!/usr/bin/env python
from broadworks_ocip import BroadworksAPI

api = BroadworksAPI(
    host="localhost",
    username="username@example.com",
    password="password",
    session_id="00000000-1111-2222-3333-444444444444",
)


def test_default_session_id():
    xml = (
        b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
        b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
        b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
        b'<command echo="" xsi:type="SystemSoftwareVersionGetResponse" xmlns="">'
        b"<version>24</version>"
        b"</command>"
        b"</BroadsoftDocument>\n"
    )
    decoded = api.decode_xml(xml)
    assert decoded.type_ == "SystemSoftwareVersionGetResponse"
    assert decoded.version == "24"
    assert decoded.session_id == "00000000-1111-2222-3333-444444444444"


def test_real_session_id():
    xml = (
        b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
        b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
        b'<sessionId xmlns="">770f845b-2caf-44bc-9291-e9eac7afaaf4</sessionId>'
        b'<command echo="" xsi:type="SystemSoftwareVersionGetResponse" xmlns="">'
        b"<version>24</version>"
        b"</command>"
        b"</BroadsoftDocument>\n"
    )
    decoded = api.decode_xml(xml)
    assert decoded.type_ == "SystemSoftwareVersionGetResponse"
    assert decoded.version == "24"
    assert decoded.session_id == "770f845b-2caf-44bc-9291-e9eac7afaaf4"



#!/usr/bin/env python
import pytest

from broadworks_ocip import BroadworksAPI
from broadworks_ocip.exceptions import OCIErrorAttributeMissing
from broadworks_ocip.exceptions import OCIErrorUnexpectedAttribute

api = BroadworksAPI(
    host="localhost",
    username="username@example.com",
    password="password",
    session_id="00000000-1111-2222-3333-444444444444",
)


class TestCommandClassConstructor:
    def test_working_base(self):
        cmd = api.get_command_object(
            "GroupAccessDeviceAddRequest14",
            service_provider_id="sp",
            group_id="grp",
            device_name="devname",
            device_type="devtype",
            protocol="SIP 2.0",
            transport_protocol="Unspecified",
            use_custom_user_name_password=True,
            access_device_credentials=api.get_type_object(
                "DeviceManagementUserNamePassword16",
                user_name="tn",
                password="tnpass",
            ),
        )
        assert cmd is not None
        assert "GroupAccessDeviceAddRequest14" in str(type(cmd))

    def test_unexpected_list(self):
        with pytest.raises(TypeError):
            api.get_command_object(
                "GroupAccessDeviceAddRequest14",
                service_provider_id="sp",
                group_id="grp",
                device_name="devname",
                device_type="devtype",
                protocol="SIP 2.0",
                transport_protocol="Unspecified",
                use_custom_user_name_password=True,
                access_device_credentials=[  # this should not be an array
                    api.get_type_object(
                        "DeviceManagementUserNamePassword16",
                        user_name="tn",
                        password="tnpass",
                    ),
                ],
            )

    def test_not_a_bool(self):
        with pytest.raises(TypeError):
            api.get_command_object(
                "GroupAccessDeviceAddRequest14",
                service_provider_id="sp",
                group_id="grp",
                device_name="devname",
                device_type="devtype",
                protocol="SIP 2.0",
                transport_protocol="Unspecified",
                use_custom_user_name_password="true",  # not a bool
                access_device_credentials=api.get_type_object(
                    "DeviceManagementUserNamePassword16",
                    user_name="tn",
                    password="tnpass",
                ),
            )

    def test_unexpected_bool(self):
        with pytest.raises(TypeError):
            api.get_command_object(
                "GroupAccessDeviceAddRequest14",
                service_provider_id="sp",
                group_id="grp",
                device_name="devname",
                device_type="devtype",
                protocol="SIP 2.0",
                transport_protocol=False,  # unexpected bool
                use_custom_user_name_password=True,
                access_device_credentials=api.get_type_object(
                    "DeviceManagementUserNamePassword16",
                    user_name="tn",
                    password="tnpass",
                ),
            )

    def test_too_many_things(self):
        with pytest.raises(TypeError):
            api.get_command_object(
                "GroupAccessDeviceAddRequest14",
                service_provider_id="sp",
                group_id="grp",
                device_name=["devname", "moredevname"],  # so many things
                device_type="devtype",
                protocol="SIP 2.0",
                transport_protocol="Unspecified",
                use_custom_user_name_password=True,
                access_device_credentials=api.get_type_object(
                    "DeviceManagementUserNamePassword16",
                    user_name="tn",
                    password="tnpass",
                ),
            )

    def test_missing_attribute(self):
        cmd = api.get_command_object(
            "GroupAccessDeviceAddRequest14",
            service_provider_id="sp",
            group_id="grp",
            device_name="devname",
            device_type="devtype",
            # protocol="SIP 2.0",   # missing optional attrobute
            transport_protocol="Unspecified",
            use_custom_user_name_password=True,
            access_device_credentials=api.get_type_object(
                "DeviceManagementUserNamePassword16",
                user_name="tn",
                password="tnpass",
            ),
        )
        assert cmd is not None

        with pytest.raises(OCIErrorAttributeMissing):
            api.get_command_object(
                "GroupAccessDeviceAddRequest14",
                # service_provider_id="sp",   # missing required attrobute
                group_id="grp",
                device_name="devname",
                device_type="devtype",
                protocol="SIP 2.0",
                transport_protocol="Unspecified",
                use_custom_user_name_password=True,
                access_device_credentials=api.get_type_object(
                    "DeviceManagementUserNamePassword16",
                    user_name="tn",
                    password="tnpass",
                ),
            )

    def test_extra_attribute(self):
        with pytest.raises(OCIErrorUnexpectedAttribute):
            api.get_command_object(
                "GroupAccessDeviceAddRequest14",
                unexpected_attribute="strange value",
                service_provider_id="sp",
                group_id="grp",
                device_name="devname",
                device_type="devtype",
                protocol="SIP 2.0",
                transport_protocol="Unspecified",
                use_custom_user_name_password=True,
                access_device_credentials=api.get_type_object(
                    "DeviceManagementUserNamePassword16",
                    user_name="tn",
                    password="tnpass",
                ),
            )

    def test_working_ex1(self):
        cmd = api.get_command_object(
            "GroupAccessDeviceAddRequest22V2",
            service_provider_id="sp",
            group_id="grp",
            device_name="devname",
            device_type="devtype",
            protocol="SIP 2.0",
            transport_protocol="Unspecified",
            use_custom_user_name_password=True,
            access_device_credentials=api.get_type_object(
                "DeviceManagementUserNamePassword16",
                user_name="tn",
                password="tnpass",
            ),
        )
        assert cmd is not None
        assert "GroupAccessDeviceAddRequest22V2" in str(type(cmd))

    def test_working_ex2(self):
        cmd = api.get_command_object(
            "GroupAccessDeviceAddRequest22V2",
            group_external_id="external_id1",
            device_external_id="external_id2",
            device_type="devtype",
            protocol="SIP 2.0",
            transport_protocol="Unspecified",
            use_custom_user_name_password=True,
            access_device_credentials=api.get_type_object(
                "DeviceManagementUserNamePassword16",
                user_name="tn",
                password="tnpass",
            ),
        )
        assert cmd is not None
        assert "GroupAccessDeviceAddRequest22V2" in str(type(cmd))

    @pytest.mark.xfail  # we do not check for this yet
    def test_exclusive_groups(self):
        with pytest.raises(OCIErrorUnexpectedAttribute):
            api.get_command_object(
                "GroupAccessDeviceAddRequest22V2",
                service_provider_id="sp",  # one group
                group_id="grp",
                device_name="devname",
                group_external_id="external_id1",  # other group - should only have one
                device_external_id="external_id2",
                device_type="devtype",
                protocol="SIP 2.0",
                transport_protocol="Unspecified",
                use_custom_user_name_password=True,
                access_device_credentials=api.get_type_object(
                    "DeviceManagementUserNamePassword16",
                    user_name="tn",
                    password="tnpass",
                ),
            )

    @pytest.mark.xfail  # we do not check for this yet
    def test_group_missing_attribute(self):
        with pytest.raises(TypeError):
            api.get_command_object(
                "GroupAccessDeviceAddRequest22V2",
                service_provider_id="sp",  # one group
                group_id="grp",
                # device_name="devname",    # this should be set
                device_type="devtype",
                protocol="SIP 2.0",
                transport_protocol="Unspecified",
                use_custom_user_name_password=True,
                access_device_credentials=api.get_type_object(
                    "DeviceManagementUserNamePassword16",
                    user_name="tn",
                    password="tnpass",
                ),
            )

    def test_department_key_type(self):
        cmd = api.get_command_object(
            "UserModifyRequest16",
            user_id="someuser@example.com",
            department=api.get_type_object(
                "EnterpriseDepartmentKey",
                service_provider_id="mysp",
                name="mydept",
            ),
        )
        assert cmd is not None
        assert "UserModifyRequest16" in str(type(cmd))



#!/usr/bin/env python
"""Tests for `broadworks_ocip` package."""
from collections import namedtuple

import pytest  # noqa: F401
from null_object import Null

from broadworks_ocip import BroadworksAPI

BASIC_API_PARAMS = {
    "host": "localhost",
    "username": "username@example.com",
    "password": "password",
    "session_id": "00000000-1111-2222-3333-444444444444",
}


class serviceProviderTable(
    namedtuple(
        "serviceProviderTable",
        ["service_provider_id", "service_provider_name", "is_enterprise"],
    ),
):
    pass


class registrationTable(
    namedtuple(
        "registrationTable",
        [
            "device_level",
            "device_name",
            "order",
            "uri",
            "expiration",
            "line_port",
            "endpoint_type",
            "public_net_address",
            "public_port",
            "private_net_address",
            "private_port",
            "user_agent",
            "lockout_started",
            "lockout_expires",
            "lockout_count",
            "access_info",
            "path_header",
            "registration_active",
        ],
    ),
):
    pass


def make_command_from_xml(xml, command, serialised):
    """Create a Broadworks XML command framgment from the argumenta"""
    api = BroadworksAPI(**BASIC_API_PARAMS)
    generated = api.decode_xml(xml)
    assert generated.type_ == command
    assert generated.to_dict() == serialised


def test_service_provider_list_xml():
    make_command_from_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command echo="" xsi:type="ServiceProviderGetListResponse" xmlns="">'
            b"<serviceProviderTable>"
            b"<colHeading>Service Provider Id</colHeading>"
            b"<colHeading>Service Provider Name</colHeading>"
            b"<colHeading>Is Enterprise</colHeading>"
            b"<row><col>one</col><col/><col>true</col></row>"
            b"<row><col>two</col><col/><col>true</col></row>"
            b"<row><col>three</col><col/><col>true</col></row>"
            b"<row><col>four</col><col/><col>true</col></row>"
            b"<row><col>five</col><col/><col>true</col></row>"
            b"<row><col>six</col><col>six name</col><col>true</col></row>"
            b"<row><col>seven</col><col/><col>true</col></row>"
            b"<row><col>eight</col><col/><col>true</col></row>"
            b"<row><col>nine</col><col/><col>true</col></row>"
            b"<row><col>ten</col><col>ten name</col><col>true</col></row>"
            b"<row><col>eleven</col><col/><col>true</col></row>"
            b"</serviceProviderTable>"
            b"</command>"
            b"</BroadsoftDocument>\n"
        ),
        "ServiceProviderGetListResponse",
        {
            "session_id": "00000000-1111-2222-3333-444444444444",
            "service_provider_table": [
                serviceProviderTable(
                    service_provider_id="one",
                    service_provider_name=None,
                    is_enterprise="true",
                ),
                serviceProviderTable(
                    service_provider_id="two",
                    service_provider_name=None,
                    is_enterprise="true",
                ),
                serviceProviderTable(
                    service_provider_id="three",
                    service_provider_name=None,
                    is_enterprise="true",
                ),
                serviceProviderTable(
                    service_provider_id="four",
                    service_provider_name=None,
                    is_enterprise="true",
                ),
                serviceProviderTable(
                    service_provider_id="five",
                    service_provider_name=None,
                    is_enterprise="true",
                ),
                serviceProviderTable(
                    service_provider_id="six",
                    service_provider_name="six name",
                    is_enterprise="true",
                ),
                serviceProviderTable(
                    service_provider_id="seven",
                    service_provider_name=None,
                    is_enterprise="true",
                ),
                serviceProviderTable(
                    service_provider_id="eight",
                    service_provider_name=None,
                    is_enterprise="true",
                ),
                serviceProviderTable(
                    service_provider_id="nine",
                    service_provider_name=None,
                    is_enterprise="true",
                ),
                serviceProviderTable(
                    service_provider_id="ten",
                    service_provider_name="ten name",
                    is_enterprise="true",
                ),
                serviceProviderTable(
                    service_provider_id="eleven",
                    service_provider_name=None,
                    is_enterprise="true",
                ),
            ],
        },
    )


def test_service_pack_list_xml():
    make_command_from_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command echo="" xsi:type="ServiceProviderServicePackGetListResponse" xmlns="">'
            b"<servicePackName>Pack_1</servicePackName>"
            b"<servicePackName>Pack_2</servicePackName>"
            b"<servicePackName>Pack_3</servicePackName>"
            b"<servicePackName>Pack_4</servicePackName>"
            b"<servicePackName>Pack_5</servicePackName>"
            b"<servicePackName>Pack_6</servicePackName>"
            b"<servicePackName>Pack_7</servicePackName>"
            b"<servicePackName>Pack_8</servicePackName>"
            b"</command>"
            b"</BroadsoftDocument>\n"
        ),
        "ServiceProviderServicePackGetListResponse",
        {
            "session_id": "00000000-1111-2222-3333-444444444444",
            "service_pack_name": [
                "Pack_1",
                "Pack_2",
                "Pack_3",
                "Pack_4",
                "Pack_5",
                "Pack_6",
                "Pack_7",
                "Pack_8",
            ],
        },
    )


def test_service_provider_service_pack_get_detail_list_response_xml():
    userServiceTable = namedtuple(
        "userServiceTable",
        ["service", "authorized", "allocated", "available"],
    )
    make_command_from_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command echo="" xsi:type="ServiceProviderServicePackGetDetailListResponse" xmlns="">'
            b"<servicePackName>Service-Pack-Standard</servicePackName>"
            b"<servicePackDescription>Service Pack - Standard</servicePackDescription>"
            b"<isAvailableForUse>true</isAvailableForUse>"
            b"<servicePackQuantity><unlimited>true</unlimited></servicePackQuantity>"
            b"<assignedQuantity><unlimited>true</unlimited></assignedQuantity>"
            b"<allowedQuantity><unlimited>true</unlimited></allowedQuantity>"
            b"<userServiceTable>"
            b"<colHeading>Service</colHeading>"
            b"<colHeading>Authorized</colHeading>"
            b"<colHeading>Allocated</colHeading>"
            b"<colHeading>Available</colHeading>"
            b"<row>"
            b"<col>Call Center - Standard</col>"
            b"<col>Unlimited</col>"
            b"<col>Unlimited</col>"
            b"<col>Unlimited</col>"
            b"</row>"
            b"</userServiceTable>"
            b"</command></BroadsoftDocument>"
        ),
        "ServiceProviderServicePackGetDetailListResponse",
        {
            "session_id": "00000000-1111-2222-3333-444444444444",
            "allowed_quantity": {"quantity": None, "unlimited": True},
            "assigned_quantity": {"quantity": None, "unlimited": True},
            "is_available_for_use": True,
            "service_pack_description": "Service Pack - Standard",
            "service_pack_name": "Service-Pack-Standard",
            "service_pack_quantity": {"quantity": None, "unlimited": True},
            "user_service_table": [
                userServiceTable(
                    service="Call Center - Standard",
                    authorized="Unlimited",
                    allocated="Unlimited",
                    available="Unlimited",
                ),
            ],
        },
    )


def test_user_call_forwarding_always_get_response():
    make_command_from_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-2"?>'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command echo="" xsi:type="UserCallForwardingAlwaysGetResponse" xmlns="">'
            b"<isActive>false</isActive>"
            b"<forwardToPhoneNumber>1234567890</forwardToPhoneNumber>"
            b"<isRingSplashActive>true</isRingSplashActive>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        "UserCallForwardingAlwaysGetResponse",
        {
            "session_id": "00000000-1111-2222-3333-444444444444",
            "is_active": False,
            "forward_to_phone_number": "1234567890",
            "is_ring_splash_active": True,
        },
    )


def test_user_assigned_services_get_list_response():
    # Unfortunately I can't use the basic function for this due to differing object locations
    xml = (
        b'<?xml version="1.0" encoding="ISO-8859-2"?>'
        b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
        b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
        b'<command echo="" xsi:type="UserAssignedServicesGetListResponse" xmlns="">'
        b"<groupServiceEntry>"
        b"<serviceName>Music On Hold</serviceName>"
        b"<isActive>true</isActive>"
        b"</groupServiceEntry>"
        b"<userServiceEntry>"
        b"<serviceName>Anonymous Call Rejection</serviceName>"
        b"<isActive>false</isActive>"
        b"</userServiceEntry>"
        b"<userServiceEntry>"
        b"<serviceName>Three-Way Call</serviceName>"
        b"<isActive>true</isActive>"
        b"</userServiceEntry>"
        b"</command>"
        b"</BroadsoftDocument>"
    )

    api = BroadworksAPI(**BASIC_API_PARAMS)
    generated = api.decode_xml(xml)
    assert generated.type_ == "UserAssignedServicesGetListResponse"

    assert (
        generated.group_service_entry[0].to_dict()
        == api.get_type_object(  # noqa: W503
            "AssignedGroupServicesEntry",
            service_name="Music On Hold",
            is_active=True,
        ).to_dict()
    )

    assert (
        generated.user_service_entry[0].to_dict()
        == api.get_type_object(  # noqa: W503
            "AssignedUserServicesEntry",
            service_name="Anonymous Call Rejection",
            is_active=False,
        ).to_dict()
    )

    assert (
        generated.user_service_entry[1].to_dict()
        == api.get_type_object(  # noqa: W503
            "AssignedUserServicesEntry",
            service_name="Three-Way Call",
            is_active=True,
        ).to_dict()
    )


def test_group_department_add_xml():
    xml = (
        b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
        b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
        b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
        b'<command xmlns="" xsi:type="GroupDepartmentAddRequest">'
        b"<serviceProviderId>mysp</serviceProviderId>"
        b"<groupId>mygroup</groupId>"
        b"<departmentName>mydept</departmentName>"
        b'<parentDepartmentKey xsi:type="GroupDepartmentKey">'
        b"<serviceProviderId>mysp</serviceProviderId>"
        b"<groupId>mygroup</groupId>"
        b"<name>test-name</name>"
        b"</parentDepartmentKey>"
        b"<callingLineIdName>clid_name</callingLineIdName>"
        b"</command>"
        b"</BroadsoftDocument>"
    )
    api = BroadworksAPI(**BASIC_API_PARAMS)
    generated = api.decode_xml(xml)
    assert generated.type_ == "GroupDepartmentAddRequest"

    assert (
        generated.parent_department_key.to_dict()
        == api.get_type_object(  # noqa: W503
            "GroupDepartmentKey",
            service_provider_id="mysp",
            group_id="mygroup",
            name="test-name",
        ).to_dict()
    )


def test_nested_elements():
    xml = (
        b'<?xml version="1.0" encoding="ISO-8859-1"?>'
        b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
        b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
        b'<command xsi:type="UserModifyRequest22" xmlns="">'
        b"<userId>user@example.com</userId>"
        b"<phoneNumber>123456789</phoneNumber>"
        b"<extension>1219</extension>"
        b'<sipAliasList xsi:nil="true"/>'
        b"<endpoint>"
        b"<trunkAddressing>"
        b'<trunkGroupDeviceEndpoint xsi:nil="true"/>'
        b"<enterpriseTrunkName>ET02</enterpriseTrunkName>"
        b'<alternateTrunkIdentity xsi:nil="true"/>'
        b"</trunkAddressing>"
        b"</endpoint>"
        b"</command>"
        b"</BroadsoftDocument>"
    )
    api = BroadworksAPI(**BASIC_API_PARAMS)
    generated = api.decode_xml(xml)
    assert generated.type_ == "UserModifyRequest22"
    assert generated.user_id == "user@example.com"
    assert generated.phone_number == "123456789"
    assert generated.sip_alias_list is Null

    # assert generated.sip_alias_list is Null
    print(generated.endpoint["trunk_addressing"])
    assert (
        generated.endpoint["trunk_addressing"].to_dict()
        == api.get_type_object(  # noqa: W503
            "TrunkAddressingMultipleContactModify",
            trunk_group_device_endpoint=Null,
            enterprise_trunk_name="ET02",
            alternate_trunk_identity=Null,
            # physical_location=None,
        ).to_dict()
    )


def test_vpn_policy_decode():
    xml = (
        b'<?xml version="1.0" encoding="ISO-8859-1"?>'
        b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
        b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
        b'<command echo="" xsi:type="EnterpriseVoiceVPNGetPolicyResponse" xmlns="">'
        b"<minExtensionLength>0</minExtensionLength>"
        b"<maxExtensionLength>0</maxExtensionLength>"
        b"<policySelection>Public</policySelection>"
        b'<digitManipulation xsi:type="EnterpriseVoiceVPNDigitManipulationOptionalValue">'
        b"<operation>Prepend</operation>"
        b"<value>253</value>"
        b"</digitManipulation>"
        b"</command>"
        b"</BroadsoftDocument>"
    )
    api = BroadworksAPI(**BASIC_API_PARAMS)
    generated = api.decode_xml(xml)
    assert generated.type_ == "EnterpriseVoiceVPNGetPolicyResponse"
    assert generated.min_extension_length == 0
    assert generated.max_extension_length == 0
    assert generated.description is None
    assert generated.route_group_id is None
    assert generated.policy_selection == "Public"
    assert len(generated.digit_manipulation) == 1
    assert (
        generated.digit_manipulation[0].to_dict()
        == api.get_type_object(  # noqa: W503
            "EnterpriseVoiceVPNDigitManipulationOptionalValue",
            operation="Prepend",
            value="253",
        ).to_dict()
    )
    assert generated.treatment_id is None


def test_unexpected_table_column_name():
    """We did not handle unexpected characters in table column names"""
    xml = (
        b'<?xml version="1.0" encoding="ISO-8859-1"?>'
        b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
        b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
        b'<command echo="" xsi:type="UserGetRegistrationListResponse" xmlns="">'
        b"<registrationTable>"
        b"<colHeading>Device Level</colHeading>"
        b"<colHeading>Device Name</colHeading>"
        b"<colHeading>Order</colHeading>"
        b"<colHeading>URI</colHeading>"
        b"<colHeading>Expiration</colHeading>"
        b"<colHeading>Line/Port</colHeading>"
        b"<colHeading>Endpoint Type</colHeading>"
        b"<colHeading>Public Net Address</colHeading>"
        b"<colHeading>Public Port</colHeading>"
        b"<colHeading>Private Net Address</colHeading>"
        b"<colHeading>Private Port</colHeading>"
        b"<colHeading>User Agent</colHeading>"
        b"<colHeading>Lockout Started</colHeading>"
        b"<colHeading>Lockout Expires</colHeading>"
        b"<colHeading>Lockout Count</colHeading>"
        b"<colHeading>Access Info</colHeading>"
        b"<colHeading>Path Header</colHeading>"
        b"<colHeading>Registration Active</colHeading>"
        b"<row>"
        b"<col>Group</col>"
        b"<col>812345678</col>"
        b"<col>1</col>"
        b"<col>sip:812345678@1.2.3.4:5060;dtg=UC1_BVE_INT_9999_TG;reg-info=39400</col>"
        b"<col>Fri Jul 08 00:01:02 BST 2022</col>"
        b"<col>812345678@voip.hawaiiantel.xxx</col>"
        b"<col>Primary</col>"
        b"<col/>"
        b"<col/>"
        b"<col/>"
        b"<col/>"
        b"<col>PolyEdge-Edge_E220-UA/8.0.0.13670_48256712b942</col>"
        b"<col/>"
        b"<col/>"
        b"<col>0</col>"
        b"<col/>"
        b"<col/>"
        b"<col/>"
        b"</row>"
        b"</registrationTable>"
        b"</command>"
        b"</BroadsoftDocument>"
    )
    api = BroadworksAPI(**BASIC_API_PARAMS)
    generated = api.decode_xml(xml)
    assert generated.type_ == "UserGetRegistrationListResponse"
    assert generated.registration_table == [
        registrationTable(
            device_level="Group",
            device_name="812345678",
            order="1",
            uri="sip:812345678@1.2.3.4:5060;dtg=UC1_BVE_INT_9999_TG;reg-info=39400",
            expiration="Fri Jul 08 00:01:02 BST 2022",
            line_port="812345678@voip.hawaiiantel.xxx",
            endpoint_type="Primary",
            public_net_address=None,
            public_port=None,
            private_net_address=None,
            private_port=None,
            user_agent="PolyEdge-Edge_E220-UA/8.0.0.13670_48256712b942",
            lockout_started=None,
            lockout_expires=None,
            lockout_count="0",
            access_info=None,
            path_header=None,
            registration_active=None,
        ),
    ]


# end



"""
Broadworks OCI-P Interface Exception Classes

Exception classes used by the API.
"""
import attr


@attr.s(slots=True, frozen=True)
class OCIError(Exception):
    """Base Exception raised by OCI operations.

    Attributes:
        message: explanation of why it went bang
        object: the thing that went bang
    """

    message: str = attr.ib()
    object = attr.ib(default=None)

    def __str__(self):
        return f"{self.__class__.__name__}({self.message})"


@attr.s(slots=True, frozen=True)
class OCIErrorResponse(OCIError):
    """
    Exception raised when an ErrorResponse is received and decoded.

    Subclass of OCIError()
    """

    pass


@attr.s(slots=True, frozen=True)
class OCIErrorTimeOut(OCIError):
    """
    Exception raised when nothing is head back from the server.

    Subclass of OCIError()
    """

    pass


@attr.s(slots=True, frozen=True)
class OCIErrorUnknown(OCIError):
    """
    Exception raised when life becomes too much for the software.

    Subclass of OCIError()
    """

    pass


@attr.s(slots=True, frozen=True)
class OCIErrorAPISetup(OCIError):
    """
    Exception raised when life becomes too much for the software.

    Subclass of OCIError()
    """

    pass


@attr.s(slots=True, frozen=True)
class OCIErrorAttributeMissing(OCIError):
    """
    Exception raised when a required attribute is missing.

    Subclass of OCIError()
    """

    pass


@attr.s(slots=True, frozen=True)
class OCIErrorUnexpectedAttribute(OCIError):
    """
    Exception raised when additional elements passed to __init__

    Subclass of OCIError()
    """

    pass


# end



#!/usr/bin/env python
"""Tests for `broadworks_ocip` package."""
from collections import namedtuple

import pytest  # noqa: F401
from lxml import etree
from null_object import Null

from broadworks_ocip import BroadworksAPI

api = BroadworksAPI(
    host="localhost",
    username="username@example.com",
    password="password",
    session_id="00000000-1111-2222-3333-444444444444",
)


def canon_xml(inxml):
    """
    we XML canonicalise/pretty print this to prevent whitespace/quote ordering
    issues and to make any diffs easier to read.
    Returned XML is as a decoded string - prettier for diffs
    """
    return etree.tostring(etree.fromstring(inxml), pretty_print=True).decode()


def check_command_xml(wanted, cmd):
    """Create a Broadworks XML command fragment from the arguments"""
    generated = cmd.build_xml_()
    assert canon_xml(generated) == canon_xml(wanted)


def compare_command_xml(wanted, command, **kwargs):
    """Create a Broadworks XML command fragment from the arguments"""
    cmd = api.get_command_object(command, **kwargs)
    check_command_xml(wanted, cmd)


def test_authentication_request_xml():
    compare_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xmlns="" xsi:type="AuthenticationRequest">'
            b"<userId>username@example.com</userId></command></BroadsoftDocument>"
        ),
        "AuthenticationRequest",
        user_id="username@example.com",
    )


def test_error_response_xml():
    compare_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command type="Error" echo="" xsi:type="c:ErrorResponse" xmlns:c="C" xmlns="">'
            b"<summary>[Error 4962] Invalid password</summary>"
            b"<summaryEnglish>[Error 4962] Invalid password</summaryEnglish>"
            b"</command></BroadsoftDocument>"
        ),
        "ErrorResponse",
        summary="[Error 4962] Invalid password",
        summary_english="[Error 4962] Invalid password",
    )


def test_group_authorization_modify_xml():
    cmd = api.get_command_object(
        "GroupServiceModifyAuthorizationListRequest",
        service_provider_id="some_enterprise",
        group_id="somegroup",
        service_pack_authorization=[
            api.get_type_object(
                "ServicePackAuthorization",
                service_pack_name="Voicemail",
                authorized_quantity=api.get_type_object(
                    "UnboundedPositiveInt",
                    unlimited=True,
                ),
            ),
            api.get_type_object(
                "ServicePackAuthorization",
                service_pack_name="Hushmail",
                authorized_quantity=api.get_type_object(
                    "UnboundedPositiveInt",
                    quantity=32,
                ),
            ),
            api.get_type_object(
                "ServicePackAuthorization",
                service_pack_name="Phone",
                unauthorized=True,
            ),
        ],
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xmlns="" xsi:type="GroupServiceModifyAuthorizationListRequest">'
            b"<serviceProviderId>some_enterprise</serviceProviderId>"
            b"<groupId>somegroup</groupId>"
            b"<servicePackAuthorization>"
            b"<servicePackName>Voicemail</servicePackName>"
            b"<authorizedQuantity><unlimited>true</unlimited></authorizedQuantity>"
            b"</servicePackAuthorization>"
            b"<servicePackAuthorization>"
            b"<servicePackName>Hushmail</servicePackName>"
            b"<authorizedQuantity><quantity>32</quantity></authorizedQuantity>"
            b"</servicePackAuthorization>"
            b"<servicePackAuthorization>"
            b"<servicePackName>Phone</servicePackName>"
            b"<unauthorized>true</unauthorized>"
            b"</servicePackAuthorization>"
            b"</command></BroadsoftDocument>"
        ),
        cmd,
    )


def test_service_provider_service_pack_get_detail_list_response_xml():
    type = namedtuple(
        "userServiceTable",
        ["service", "authorized", "allocated", "available"],
    )
    cmd = api.get_command_object(
        "ServiceProviderServicePackGetDetailListResponse",
        service_pack_name="Service-Pack-Standard",
        service_pack_description="Service Pack - Standard",
        is_available_for_use=True,
        service_pack_quantity=api.get_type_object(
            "UnboundedPositiveInt",
            unlimited=True,
        ),
        assigned_quantity=api.get_type_object(
            "UnboundedNonNegativeInt",
            unlimited=True,
        ),
        allowed_quantity=api.get_type_object("UnboundedPositiveInt", unlimited=True),
        user_service_table=[
            type(
                service="Call Center - Standard",
                authorized="Unlimited",
                allocated="Unlimited",
                available="Unlimited",
            ),
        ],
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command echo="" xsi:type="ServiceProviderServicePackGetDetailListResponse" xmlns="">'
            b"<servicePackName>Service-Pack-Standard</servicePackName>"
            b"<servicePackDescription>Service Pack - Standard</servicePackDescription>"
            b"<isAvailableForUse>true</isAvailableForUse>"
            b"<servicePackQuantity><unlimited>true</unlimited></servicePackQuantity>"
            b"<assignedQuantity><unlimited>true</unlimited></assignedQuantity>"
            b"<allowedQuantity><unlimited>true</unlimited></allowedQuantity>"
            b"<userServiceTable>"
            b"<colHeading>Service</colHeading>"
            b"<colHeading>Authorized</colHeading>"
            b"<colHeading>Allocated</colHeading>"
            b"<colHeading>Available</colHeading>"
            b"<row>"
            b"<col>Call Center - Standard</col>"
            b"<col>Unlimited</col>"
            b"<col>Unlimited</col>"
            b"<col>Unlimited</col>"
            b"</row>"
            b"</userServiceTable>"
            b"</command></BroadsoftDocument>"
        ),
        cmd,
    )


def test_group_outgoing_calling_plan_redirecting_modify_list_request_xml():
    cmd = api.get_command_object(
        "GroupOutgoingCallingPlanRedirectingModifyListRequest",
        service_provider_id="SP01",
        group_id="GRP01",
        group_permissions=api.get_type_object(
            "OutgoingCallingPlanRedirectingPermissionsModify",
            group=True,
            international=False,
        ),
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xsi:type="GroupOutgoingCallingPlanRedirectingModifyListRequest" xmlns="">'
            b"<serviceProviderId>SP01</serviceProviderId>"
            b"<groupId>GRP01</groupId>"
            b"<groupPermissions>"
            b"<group>true</group>"
            b"<international>false</international>"
            b"</groupPermissions>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        cmd,
    )


def test_group_dn_unassign_list_request_xml():
    cmd = api.get_command_object(
        "GroupDnUnassignListRequest",
        service_provider_id="SP01",
        group_id="GRP01",
        phone_number=["+1-55512340", "+1-55512342"],
        dn_range=[
            api.get_type_object(
                "DNRange",
                min_phone_number="+1-55512345",
                max_phone_number="+1-55512350",
            ),
            api.get_type_object(
                "DNRange",
                min_phone_number="+1-55512355",
                max_phone_number="+1-55512360",
            ),
        ],
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xsi:type="GroupDnUnassignListRequest" xmlns="">'
            b"<serviceProviderId>SP01</serviceProviderId>"
            b"<groupId>GRP01</groupId>"
            b"<phoneNumber>+1-55512340</phoneNumber>"
            b"<phoneNumber>+1-55512342</phoneNumber>"
            b"<dnRange>"
            b"<minPhoneNumber>+1-55512345</minPhoneNumber>"
            b"<maxPhoneNumber>+1-55512350</maxPhoneNumber>"
            b"</dnRange>"
            b"<dnRange>"
            b"<minPhoneNumber>+1-55512355</minPhoneNumber>"
            b"<maxPhoneNumber>+1-55512360</maxPhoneNumber>"
            b"</dnRange>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        cmd,
    )


def test_service_provider_dn_delete_list_request_xml():
    cmd = api.get_command_object(
        "ServiceProviderDnDeleteListRequest",
        service_provider_id="SP01",
        phone_number=["+1-55512340", "+1-55512342"],
        dn_range=[
            api.get_type_object(
                "DNRange",
                min_phone_number="+1-55512345",
                max_phone_number="+1-55512350",
            ),
            api.get_type_object(
                "DNRange",
                min_phone_number="+1-55512355",
                max_phone_number="+1-55512360",
            ),
        ],
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xsi:type="ServiceProviderDnDeleteListRequest" xmlns="">'
            b"<serviceProviderId>SP01</serviceProviderId>"
            b"<phoneNumber>+1-55512340</phoneNumber>"
            b"<phoneNumber>+1-55512342</phoneNumber>"
            b"<dnRange>"
            b"<minPhoneNumber>+1-55512345</minPhoneNumber>"
            b"<maxPhoneNumber>+1-55512350</maxPhoneNumber>"
            b"</dnRange>"
            b"<dnRange>"
            b"<minPhoneNumber>+1-55512355</minPhoneNumber>"
            b"<maxPhoneNumber>+1-55512360</maxPhoneNumber>"
            b"</dnRange>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        cmd,
    )


def test_user_voice_messaging_user_modify_voice_management_request_xml():
    cmd = api.get_command_object(
        "UserVoiceMessagingUserModifyVoiceManagementRequest",
        user_id="fred.flintstone@boulder.org",
        is_active=True,
        send_carbon_copy_voice_message=True,
        busy_redirect_to_voice_mail=True,
        no_answer_redirect_to_voice_mail=True,
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xsi:type="UserVoiceMessagingUserModifyVoiceManagementRequest" xmlns="">'
            b"<userId>fred.flintstone@boulder.org</userId>"
            b"<isActive>true</isActive>"
            b"<sendCarbonCopyVoiceMessage>true</sendCarbonCopyVoiceMessage>"
            b"<busyRedirectToVoiceMail>true</busyRedirectToVoiceMail>"
            b"<noAnswerRedirectToVoiceMail>true</noAnswerRedirectToVoiceMail>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        cmd,
    )


def test_user_busy_lap_field_modify_request_xml():
    cmd = api.get_command_object(
        "UserBusyLampFieldModifyRequest",
        user_id="fred.flintstone@boulder.org",
        list_uri="list_sip_uri@boulder.org",
        monitored_user_id_list=api.get_type_object(
            "ReplacementUserIdList",
            user_id=["one@boulder.org", "two@boulder.org"],
        ),
        enable_call_park_notification=True,
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xmlns="" xsi:type="UserBusyLampFieldModifyRequest">'
            b"<userId>fred.flintstone@boulder.org</userId>"
            b"<listURI>list_sip_uri@boulder.org</listURI>"
            b"<monitoredUserIdList>"
            b"<userId>one@boulder.org</userId>"
            b"<userId>two@boulder.org</userId>"
            b"</monitoredUserIdList>"
            b"<enableCallParkNotification>true</enableCallParkNotification>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        cmd,
    )


def test_user_modify_department_key_xml():
    cmd = api.get_command_object(
        "UserModifyRequest16",
        user_id="fred.flintstone@boulder.org",
        department=api.get_type_object(
            "EnterpriseDepartmentKey",
            service_provider_id="mysp",
            name="mydept",
        ),
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xmlns="" xsi:type="UserModifyRequest16">'
            b"<userId>fred.flintstone@boulder.org</userId>"
            b'<department xsi:type="EnterpriseDepartmentKey">'
            b"<serviceProviderId>mysp</serviceProviderId>"
            b"<name>mydept</name>"
            b"</department>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        cmd,
    )


def test_group_department_add_xml():
    cmd = api.get_command_object(
        "GroupDepartmentAddRequest",
        service_provider_id="mysp",
        group_id="mygroup",
        department_name="mydept",
        parent_department_key=api.get_type_object(
            "GroupDepartmentKey",
            service_provider_id="mysp",
            group_id="mygroup",
            name="test-name",
        ),
        calling_line_id_name="clid_name",
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>\n'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xmlns="" xsi:type="GroupDepartmentAddRequest">'
            b"<serviceProviderId>mysp</serviceProviderId>"
            b"<groupId>mygroup</groupId>"
            b"<departmentName>mydept</departmentName>"
            b'<parentDepartmentKey xsi:type="GroupDepartmentKey">'
            b"<serviceProviderId>mysp</serviceProviderId>"
            b"<groupId>mygroup</groupId>"
            b"<name>test-name</name>"
            b"</parentDepartmentKey>"
            b"<callingLineIdName>clid_name</callingLineIdName>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        cmd,
    )


def test_nested_elements():
    cmd = api.get_command_object(
        "UserModifyRequest22",
        user_id="user@example.com",
        phone_number="123456789",
        extension="1219",
        sip_alias_list=Null,
        endpoint=dict(
            trunk_addressing=api.get_type_object(
                "TrunkAddressingMultipleContactModify",
                trunk_group_device_endpoint=Null,
                enterprise_trunk_name="ET02",
                alternate_trunk_identity=Null,
                physical_location=Null,
            ),
        ),
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xsi:type="UserModifyRequest22" xmlns="">'
            b"<userId>user@example.com</userId>"
            b"<phoneNumber>123456789</phoneNumber>"
            b"<extension>1219</extension>"
            b'<sipAliasList xsi:nil="true"/>'
            b"<endpoint>"
            b"<trunkAddressing>"
            b'<trunkGroupDeviceEndpoint xsi:nil="true"/>'
            b"<enterpriseTrunkName>ET02</enterpriseTrunkName>"
            b'<alternateTrunkIdentity xsi:nil="true"/>'
            b'<physicalLocation xsi:nil="true"/>'
            b"</trunkAddressing>"
            b"</endpoint>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        cmd,
    )


def test_nested_elements2():
    cmd = api.get_command_object(
        "UserModifyRequest22",
        user_id="user@example.com",
        phone_number="123456789",
        extension="1219",
        sip_alias_list=api.get_type_object("ReplacementSIPAliasList", sip_alias=[]),
        endpoint=dict(
            trunk_addressing=api.get_type_object(
                "TrunkAddressingMultipleContactModify",
                trunk_group_device_endpoint=Null,
                enterprise_trunk_name="ET02",
                alternate_trunk_identity=Null,
            ),
        ),
    )
    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xsi:type="UserModifyRequest22" xmlns="">'
            b"<userId>user@example.com</userId>"
            b"<phoneNumber>123456789</phoneNumber>"
            b"<extension>1219</extension>"
            b'<sipAliasList xsi:nil="true"/>'
            b"<endpoint>"
            b"<trunkAddressing>"
            b'<trunkGroupDeviceEndpoint xsi:nil="true"/>'
            b"<enterpriseTrunkName>ET02</enterpriseTrunkName>"
            b'<alternateTrunkIdentity xsi:nil="true"/>'
            b"</trunkAddressing>"
            b"</endpoint>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        cmd,
    )


def test_user_get_list_in_system():
    cmd = api.get_command_object(
        "UserGetListInSystemRequest",
        response_size_limit=10,
        search_criteria_dn=[
            api.get_type_object(
                "SearchCriteriaDn",
                mode="Contains",
                value="12345678",
                is_case_insensitive=True,
            ),
        ],
    )
    assert cmd is not None
    assert "UserGetListInSystemRequest" in str(type(cmd))

    check_command_xml(
        (
            b'<?xml version="1.0" encoding="ISO-8859-1"?>'
            b'<BroadsoftDocument protocol="OCI" xmlns="C" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b'<sessionId xmlns="">00000000-1111-2222-3333-444444444444</sessionId>'
            b'<command xsi:type="UserGetListInSystemRequest" xmlns="">'
            b"<responseSizeLimit>10</responseSizeLimit>"
            b"<searchCriteriaDn>"
            b"<mode>Contains</mode>"
            b"<value>12345678</value>"
            b"<isCaseInsensitive>true</isCaseInsensitive>"
            b"</searchCriteriaDn>"
            b"</command>"
            b"</BroadsoftDocument>"
        ),
        cmd,
    )


# end
